-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
--                                                                         --
--            MODULE OBJ_parser.hs    -     December 2002                  --
--                                                                         --
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------

module OBJ_parser where

import Parsec
import List

--Main function for parsing a OBJ program
parseOBJ :: String -> IO OBJProgram
parseOBJ filename  
   = do { resultparse <- parseFromFile object filename
       ; case resultparse of
           Left err -> do{ putStr "parse error at "
                         ; print err
                         ; return (Obj "error" [] [] [] [])
                         }
           Right x  -> return (collectConstructors (fixVariables (fixCObjVars x)))
       }

--Post-process for transforming variables with types of CafeOBJ into variables
fixCObjVars :: OBJProgram -> OBJProgram
fixCObjVars (Obj f sD oD vD eD) = (Obj f sD oD (vD++nub (extract eD)) (map remove eD))
  where extract [] = []
        extract (Eq lhs _:ts) = extractVars lhs ++ extract ts
        extractVars (Call v []) = 
          if (elem ':' v) 
          then [(takeWhile (':'/=) v,tail (dropWhile (':'/=) v))] else []
        extractVars (Call _ ts) = concatMap extractVars ts
        remove (Eq lhs rhs) = Eq (removeVars lhs) rhs
        removeVars (Call v ts) = Call (takeWhile (':'/=) v) (map removeVars ts)

--Post-process for transforming variables identified as terms into variables
fixVariables :: OBJProgram -> OBJProgram
fixVariables (Obj f sD oD vD eD) = (Obj f sD oD vD (fixV (packV vD) eD))
  where packV l = map (\(x,y)->x) l
        fixV vsV eqs = map (fixE vsV) eqs 
        fixE vsE (Eq lhs rhs) = Eq (fixT vsE lhs) (fixT vsE rhs)
        fixT vsT (Call v []) = if (v `elem` vsT) then (Var v) else (Call v [])
        fixT vsT (Call v ts) = Call v (map (fixT vsT) ts)

--Post-process for transforming variables identified as terms into variables
collectConstructors (Obj f sD oD vD eD) = (Obj f sD (map fixOp oD) vD eD)
  where fixOp (Op f _ sort stra) = Op f (not (f `elem`defs)) sort stra
        defs = nub (map def eD)
        def (Eq (Call f _) _) = f

--BNF grammar for OBJ programs--------------------------------------

--Data types for an OBJ program
data OBJProgram = Obj String [SortD] [OpD] [VarD] [EqD]
                deriving (Show,Eq)
type SortD = String
data OpD = Op OpID Bool [SortD] StratD -- Op name constructor type strategy
         deriving (Show,Eq)
type OpID = String
type StratD = [Int]
type VarD = (VarID,SortD)
type VarID = String
data EqD = Eq TermD TermD
         deriving (Show,Eq)
type TermD = Term String

--General Data type
data Term sym = Var sym
              | Call sym [Term sym]
         deriving (Show,Eq)

object = objectOBJ <|> objectCafeOBJ

--------------------------OBJ Syntax-------------------------
objectOBJ :: Parser OBJProgram
objectOBJ = 
  do{ spaces
    ; string "obj" 
    ; spaces1
    ; name <- modID
    ; spaces1
    ; string "is"
    ; spaces
    ; (sorts,ops,vars,eqs) <- modEltOBJ
    ; spaces
    ; return (Obj name sorts ops vars eqs)
  }

modEltOBJ :: Parser ([SortD],[OpD],[VarD],[EqD])

modEltOBJ = do{ spaces
           ;    do{ char 'e'
                  ;     do{ string "ndo"
                          ; return ([],[],[],[])
                          }
                    <|> do{ p1 <- eqDecl
                          ; p2 <- rest p1
                          ; return p2
                          }
                  }
            <|> do{ p1 <- sortsDecl
                  ; p2 <- rest p1
                  ; return p2
                  }

            <|> do{ p1 <- opDecl
                  ; p2 <- rest p1
                  ; return p2
                  }

            <|> do{ p1 <- varDecl
                  ; p2 <- rest p1
                  ; return p2
                  }

            <|> do{ char 'e'
                  ; p1 <- eqDecl
                  ; p2 <- rest p1
                  ; return p2
                  }
           }
  where rest (sorts,ops,vars,eqs) = 
                do{ (sortL,opL,varL,eqL) <- modEltOBJ
                  ; return (sorts++sortL,ops++opL,vars++varL,eqs++eqL)
                  }

--sortsDecl :: Parser ([SortD],[OpD],[VarD],[EqD])
sortsDecl = do{ do{string "sort"; char 's' <|> space}
              ; spaces
              ; sorts <- sortID `sepBy1` spaces1
              ; spaces1
              ; char '.'
              ; return (sorts,[],[],[])
              }

--opDecl :: Parser ([SortD],[OpD],[VarD],[EqD])
opDecl = do{ string "op" 
           ; spaces1
           ; op <- opID
           ; spaces1 ; char ':' ; spaces1
           ; sorts <- typeOf
           ; spaces1
           ; strat <- strategy
           ; return ([],[Op op False sorts strat],[],[])
           }
   where typeOf = do{ sortsArg <- sortID `sepBy` spaces1
                    ; spaces
                    ; string "->"
                    ; spaces
                    ; sortFinal <- sortID
                    ; return (sortsArg++[sortFinal])
                    }
         strategy = do{ char '['
                      ; string "strat"
                      ; spaces1
                      ; char '('
                      ; strat <- integer `sepBy` space
                      ; char ')'
                      ; spaces
                      ; char ']'
                      ; spaces1
                      ; char '.'
                      ; return strat
                      }
                <|> do{ char '.'
                      ; return []
                      }


--varDecl :: Parser ([SortD],[OpD],[VarD],[EqD])
varDecl = do{ do{string "var"; char 's' <|> space}
            ; spaces
            ; vars <- varID `sepBy1` spaces1
            ; spaces1
            ; char ':'
            ; spaces1
            ; sort <- sortID
            ; spaces1
            ; char '.'
            ; return ([],[],map (\x->(x,sort)) vars,[])
            }

--eqDecl :: Parser ([SortD],[OpD],[VarD],[EqD])
eqDecl = do{ string "q"
           ; spaces1
           ; lhs <- term
           ; spaces
           ; char '='
           ; spaces
           ; rhs <- term
           ; spaces1
           ; char '.'
           ; return ([],[],[],[Eq lhs rhs])
           }

term :: Parser TermD
term = do{ spacesN
         ; str1 <- opID
         ; t <-  do{ char '('
                   ; spacesN
                   ; ts <- do{spacesN;te<-term;spacesN;return te} `sepBy1` (char ',')
                   ; spacesN
                   ; char ')'
                   ; return (Call str1 ts)
                   }
           <|> return (Call str1 [])
         ; return t
         }

--------------------------CafeOBJ Syntax-------------------------
objectCafeOBJ :: Parser OBJProgram
objectCafeOBJ = 
  do{ spaces
    ; string "mod!" 
    ; spaces1
    ; name <- modID
    ; spaces1
    ; string "{"
    ; spaces
    ; (sorts,ops,vars,eqs) <- modEltCOBJ
    ; spaces
    ; return (Obj name sorts ops vars eqs)
  }

modEltCOBJ :: Parser ([SortD],[OpD],[VarD],[EqD])

modEltCOBJ = do{ spaces
           ;    do{ string "}"
                  ; return ([],[],[],[])
                  }

            <|> do{ p1 <- sortsDeclCOBJ
                  ; p2 <- rest p1
                  ; return p2
                  }

            <|> do{ p1 <- opDeclCOBJ
                  ; p2 <- rest p1
                  ; return p2
                  }

            <|> do{ p1 <- varDecl
                  ; p2 <- rest p1
                  ; return p2
                  }

            <|> do{ p1 <- eqDeclCOBJ
                  ; p2 <- rest p1
                  ; return p2
                  }
           }
  where rest (sorts,ops,vars,eqs) = 
                do{ (sortL,opL,varL,eqL) <- modEltCOBJ
                  ; return (sorts++sortL,ops++opL,vars++varL,eqs++eqL)
                  }

--sortsDecl :: Parser ([SortD],[OpD],[VarD],[EqD])
sortsDeclCOBJ = 
   do{ char '['
     ; spaces
     ; sorts <- sortID `sepBy1` (char ',')
     ; spaces
     ; char ']'
     ; return (sorts,[],[],[])
   }

--opDecl :: Parser ([SortD],[OpD],[VarD],[EqD])
opDeclCOBJ = 
   do{ string "op" 
     ; spaces1
     ; op <- opID
     ; spaces1 ; char ':' ; spaces1
     ; sorts <- typeOf
     ; spaces1
     ; strat <- strategy
     ; return ([],[Op op False sorts strat],[],[])
   }
  where typeOf = 
          do{ sortsArg <- sortID `sepBy` spaces1
            ; spaces
            ; string "->"
            ; spaces
            ; sortFinal <- sortID
            ; return (sortsArg++[sortFinal])
          }
        strategy = 
          do{ char '{'
            ; spaces
            ; string "constr strat:"
            ; spaces1
            ; char '('
            ; strat <- integer `sepBy` space
            ; char ')'
            ; spaces
            ; char '}'
            ; spaces1
            ; char '.'
            ; return strat
          }
          <|> 
          do{ char '.'
            ; return []
          }

--eqDecl :: Parser ([SortD],[OpD],[VarD],[EqD])
eqDeclCOBJ = 
   do{ string "eq"
     ; spaces1
     ; lhs <- term
     ; spaces
     ; char '='
     ; spaces
     ; rhs <- term
     ; spaces1
     ; char '.'
     ; return ([],[],[],[Eq lhs rhs])
   }

-- Auxiliary parser operators
modID :: Parser String
modID = identifier upper (alphaNum <|> (oneOf "_-'"))

sortID :: Parser String
sortID = identifier upper (alphaNum <|> (oneOf "_-'"))

opID :: Parser String
opID = identifier alphaNum (alphaNum <|> (oneOf "_':-+")) 

varID :: Parser String
varID = identifier upper (alphaNum <|> (oneOf "_-'"))

-- <identifier pFirst PLast> parses an identifier where the first symbol
-- and the rest have different parsers
identifier :: Parser Char -> Parser Char -> Parser String
identifier pFirst pLast
  = do{ c <- pFirst
      ; cs <- many pLast
      ; return (c:cs)
      }

spacesN = skipMany (oneOf "\r\n" <|> space)

spaces1N = skipMany1 (space <|> oneOf "\r\n")

spaces1 = skipMany1 space 

natural 
  = do{ cs <- many1 digit
      ; return (str2int 0 (map digitToInt2 cs))
      }
  where str2int n [] = n
        str2int n (c:cs) = str2int ((n*10)+c) cs

integer = natural
      <|> do{ char '+'
            ; natural
            }
      <|> do{ char '-'
            ; i <- natural
            ; return (-1*i)
            }
 
--- Functions copied from Hugs Prelude

-- Digit conversion operations
digitToInt2 :: Char -> Int
digitToInt2 c
  | isDigit2 c           =  fromEnum c - fromEnum '0'
  | c >= 'a' && c <= 'f' =  fromEnum c - fromEnum 'a' + 10
  | c >= 'A' && c <= 'F' =  fromEnum c - fromEnum 'A' + 10
  | otherwise            =  error "Char.digitToInt: not a digit"


isDigit2 c              =  c >= '0'   &&  c <= '9'
