-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
--                                                                         --
--                 MODULE Terms.hs    -     December 2002                  --
--                                                                         --
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------

module Terms where

import Utilities

--
--  Signatures and variables
--

type IdVariables = String
type IdFunctions = String
type Sorts = String
type RepArity id = [(id,Int)]
--type Arity id = id -> Int
type RepFunctionality id = [(id,[Sorts])]
--type Functionality id = id -> [Sorts] -- Return sort is the first one
type Signature id = ([id],[id],RepArity id,RepFunctionality id)

--
--  Terms and substitutions
--

data Terms id = V id | W id | C id | F id [Terms id] deriving (Eq,Show)
type Substitution id = id -> Terms id

--
--  Rewrite rules and TRSs
--

data RewRule id = Terms id :-> Terms id deriving (Eq,Show)
-- data Equation = Terms := Terms 
type Equation = RewRule 

type TRS id = (Signature id,[RewRule id])
--type MuTRS id = (MuSignature id,[RewRule id])

--
--  Rewrite rules and TRSs
--

type EqSpec id = TRS id
type DebProblem id = (TRS id,EqSpec id)

--------------------------------------------------------------
--
--  Functions on terms
--
--------------------------------------------------------------

--
-- mapTerms :: (Terms -> Terms) -> Terms -> Terms
--
--
--

mapTerms h (F f ts) = h (F f (map (mapTerms h) ts))
mapTerms h t = h t

--
--  (+!+) :: [a] -> [a] -> [a]
--
-- (xs +!+ ys) concatenates the lists xs and ys and removes duplicated
-- items
--

removeDuplicates [] = []
removeDuplicates (x:xs) = x:removeDuplicates (filter (/=x) xs)

xs +!+ ys = removeDuplicates (xs ++ ys)

--
--  isVarTerm::Terms -> Bool
--
--  (isVarTerm t) returns True if t is a variable term

isVarTerm (V _) = True
isVarTerm (W _) = True
isVarTerm _ = False

--
--
--

varsTerm (C _) = []
varsTerm (V x) = [x]
varsTerm (W x) = [x]
varsTerm (F f ts) = removeDuplicates (concat (map varsTerm ts))

--
--  rootOfTerm::Terms -> String
--
--  (rootOfTerm t) returns the root symbol of a term

rootOfTerm (V x) = x
rootOfTerm (W x) = x
rootOfTerm (C c) = c
rootOfTerm (F f _) = f


--
-- subTerm::Terms -> Position -> Terms
--

subTerm t [] = t
subTerm (F f ts) (i:p) = subTerm (ts !! (i-1)) p

--
-- replaceSubTerm::Terms -> Position -> Terms -> Terms
--

replaceSubTerm t [] s = s
replaceSubTerm (F f ts) (i:p) s = F f (replaceL ts i (replaceSubTerm (ts !! (i-1)) p s))

replaceL (x:xs) 1 y = y:xs
replaceL (x:xs) n y = x:(replaceL xs (n-1) y)

--
-- applySubst::Substitution -> Terms -> Terms
--

applySubst subst (C c) = C c
applySubst subst (V x) = subst x
applySubst subst (F f ts) = F f (map (applySubst subst) ts)


--
-- makeListOfNewVariables :: Int -> [Strings]
--
-- (makeListOfNewVariables n) creates a string of variables named "X1",...,"Xn"
-- 

makeListOfNewVariables 1 = ["X"]
makeListOfNewVariables n = map (('X':).show) [1..n]

--
-- makeNewVariables :: Int -> [Terms]
--
-- (makeNewVariables n) creates a list of variable terms labelled "X1",...,"Xn"
-- 

makeNewVariables n = map V (makeListOfNewVariables n)

--
-- dupName :: String -> String
--
-- (dupName xs) just inserts two initial symbols
--

dupName xs = 'n':'_':'_':xs
unDupName ('n':'_':'_':name) = name

--------------------------------------------------------------
--
--  Syntactic analysis
--
--------------------------------------------------------------

sepList = ['(',')',',']

--
-- nonSep::Char -> Bool
--

nonSep x = not (elem x sepList)

--
-- removeBlanks::String -> String
--

removeBlanks::String -> String
removeBlanks = filter (not . isSpace)

isIdentifier "" = True
isIdentifier (x:xs) | isDigit x || isAlpha x || x=='_' = isIdentifier xs
isIdentifier _ = False

--
-- isVariable::String -> Bool
--
-- Checks whether a string is a valid function identifier
--

isVariable (x:xs) | isAlpha x && isUpper x = isIdentifier xs
                  | otherwise = False
                  
--
-- isFunction::String -> Bool
--
-- Checks whether a string is a valid function identifier
--

isFunction "" = False
isFunction ('_':xs) = isFunction xs
isFunction (x:xs) | isDigit x = isIdentifier xs
                  | isAlpha x && isLower x = isIdentifier xs
                  | otherwise = False

--
-- stringToTerm::String -> Terms
--

stringToTerm = parse . (getTokens [])                  

--
-- parse::[String] -> Terms
--
-- From token list into a term.
--

parse [id] | isVariable id = V id
           | isFunction id = C id
            
parse (f:"(":xs) | last xs == ")" && isFunction f = F f (parseArgs [] 0 (init xs)) 

parseArgs arg n [] = [parse (reverse arg)]
parseArgs arg 0 (",":xs) = [parse (reverse arg)]++parseArgs [] 0 xs
parseArgs arg n ("(":xs) = parseArgs ("(":arg) (n+1) xs 
parseArgs arg n (")":xs) = parseArgs (")":arg) (n-1) xs
parseArgs arg n (x:xs) = parseArgs (x:arg) n xs

--
-- getTokens::[String] -> String -> [String]
--

getTokens tokenList [] = tokenList
getTokens tokenList cad = tokenList++(token:getTokens [] others)
   where id    = takeWhile nonSep cad
         cad'  = dropWhile nonSep cad
         token | id=="" = [head cad']
               | otherwise = id
         others | id=="" = tail cad'
                | otherwise = cad'


--
-- termToString::Terms -> String
--

markCutVar x = x++"_"

termToString (C c) = c
termToString (V x) = x
termToString (W x) = markCutVar x
termToString (F f ts) = f++('(':foldl1 (+++) (map termToString ts))++")"
                    where xs +++ ys = xs++(',':ys)



--
--
-- Obtaining a signature
--
--

extractSignatureTerm (functions,variables,repArity) (V x) = (functions,variables+!+[x],repArity)
extractSignatureTerm (functions,variables,repArity) (C c) = (functions+!+[c],variables,(c,0):repArity)
extractSignatureTerm (functions,variables,repArity) (F f ts) = (functions',variables',repArity')
   where functions' = functions+!+[f]+!+functionsInSubterms
         variables' = variables+!+variablesInSubterms
         repArity'  = (f,length ts):repArity++repAritiesInSubterms
         (functionsInSubterms,variablesInSubterms,repAritiesInSubterms) =
            mixComponents (map (extractSignatureTerm ([],[],[])) ts)
         mixComponents [(x,y,z)] = (x,y,z)
         mixComponents ((x,y,z):tuples) = let (xs,ys,zs) = mixComponents tuples in (x+!+xs,y+!+ys,z++zs)
         
extractSignature sig [] = sig
extractSignature sig ((l:->r):rules) = extractSignature sig' rules 
   where (funcs,vars,repArity) = extractSignatureTerm sig l 
         (funcs',vars',repArity') = extractSignatureTerm sig r 
         sig' = (funcs+!+funcs',vars+!+vars',repArity++repArity')

--
--
-- Parsing of a TRS in simple syntax
--
--

parseTRS [] = []
parseTRS (rule:rules) = ((stringToTerm l) :-> (stringToTerm r)) : parseTRS rules
   where (l,c:r') = break (\x -> (x=='-' || x=='=')) rule
         r = if c=='-' then tail r' else r'

obtainTRS fileTRS = ((fs,vars,repArity,repFunctionality),rules)
   where rules = parseTRS (words (filter (/=' ') fileTRS))
         (fs,vars,repArity) = extractSignature ([],[],[]) rules
         ar = toFunc repArity
         repFunctionality = [(f,"S":replicate (ar f) "S")| f<-fs]

--
--
-- Formatting functions
--
--

formatFunctions functions = ('\n':'\n':"Functions: ")++(foldr (++) [] (map (' ':) functions))

formatVariables variables = ('\n':'\n':"Variables: ")++(foldr (++) [] (map (' ':) variables))

formatArities functions repArity = 
   ('\n':'\n':"Arities: ")++
   ('\n':(foldr (++) [] ['\n':'(':(f++',':(show (ar f))++")") | f<-functions]))
   where ar = toFunc repArity
   
formatFunctionalities functions repFuncty = 
   ('\n':'\n':"Functionalities: ")++
   '\n':(foldr (++) [] ['\n':(f++' ':':':(formatFunc f)) | f<-functions])
   where 
     functy = toFunc repFuncty
     formatFunc f = concat (map (' ':) (tail (functy f)))++' ':'-':'>':' ':head (functy f)


--
-- showFunctions :: [Rules] -> String
--
-- (showFunctions functions) formats a list of rules as a string (one
-- line per rule)
--

showFunctions = putStrLn . formatFunctions

--
-- showVariables :: [Rules] -> String
--
-- (showVariables functions) formats a list of rules as a string (one
-- line per rule)
--

showVariables = putStrLn . formatVariables

--
-- showArity :: [Rules] -> String
--
-- (showArity functions arity) formats a list of rules as a string (one
-- line per rule)
--

showArity functions = putStrLn . (formatArities functions)

--
-- showSignature :: [Rules] -> String
--
-- (showArity functions arity) formats a list of rules as a string (one
-- line per rule)
--

showSignature (functions,variables,arity,functionality) = 
   putStrLn ((formatFunctions functions)++
            (formatVariables variables)++
            (formatArities functions arity)++
            (formatFunctionalities functions functionality))

--
-- showRules :: [Rules] -> String
--
-- (showRules rules) formats a list of rules as a string (one
-- line per rule)
--

joinRules rules = foldr (++) [] [(termToString l)++" -> "++(termToString r)++['\n'] | l:->r <- rules]
   
formatRules rules = ('\n':'\n':"Rules: ")++('\n':'\n':(joinRules rules))

formatRulesNoTitle rules = '\n':(joinRules rules)

showRules = putStrLn . formatRulesNoTitle

showTRS ((functions,variables,arity,functionality),rules) = 
   putStrLn ((formatFunctions functions)++(formatVariables variables)++
            (formatArities functions arity)++
            (formatFunctionalities functions functionality)++
            (formatRules rules))
            
--
-- showEquations :: [Equations] -> String
--
-- (showEquations eqs) formats a list of rules as a string (one
-- line per rule)
--

joinEqs eqs = foldr (++) [] [(termToString l)++" = "++(termToString r)++['\n'] | l:->r <- eqs]
   
formatEqs eqs = ('\n':'\n':"Equations: ")++('\n':'\n':(joinEqs eqs))

formatEqsNoTitle eqs = '\n':(joinEqs eqs)

showEquations = putStrLn . formatEqsNoTitle
            
showGroupEquations groupEqs = putStrLn (concat [formatEqs eqs | eqs <- groupEqs])



--
--
--
--
--
--

isSymbolInTerm s (V x) = x==s
isSymbolInTerm s (W x) = x==s
isSymbolInTerm s (C c) = c==s
isSymbolInTerm s (F f ts) = 
   if s==f then True else or (map (isSymbolInTerm s) ts)
isSymbolInTerm s _ = False

isVarInTerm (V x) (V y) = x==y
isVarInTerm (W x) (W y) = x==y
isVarInTerm x (F f ts) = or (map (isVarInTerm x) ts)
isVarInTerm _ _ = False

--
--
-- Substitutions as equations
--
--

composeSubst e [] = [e]
composeSubst e (((V x) :-> s):subst) = ((V x) :-> applyToTerm e s):composeSubst e subst
composeSubst e (((W x) :-> s):subst) = ((W x) :-> applyToTerm e s):composeSubst e subst

applyToTerm ((V x) :-> s) (V y) 
  | x == y = s
  | otherwise = V y
applyToTerm ((W x) :-> s) (W y) 
  | x == y = s
  | otherwise = W y
applyToTerm e (F f ts) = F f (map (applyToTerm e) ts)
applyToTerm _ t = t

applyToEq e (t :-> s) = (applyToTerm e t) :-> (applyToTerm e s)

applySubstEqToTerm [] t = t
applySubstEqToTerm (((V x) :-> s):subst) t@(V y)
  | x == y = s
  | otherwise = applySubstEqToTerm subst t
applySubstEqToTerm (((W x) :-> s):subst) t@(W y)
  | x == y = s
  | otherwise = applySubstEqToTerm subst t
applySubstEqToTerm subst (F f ts) = F f (map (applySubstEqToTerm subst) ts)
applySubstEqToTerm _ t = t

--
--
--
--
--

removeEqsFrom [] eqs = eqs
removeEqsFrom eqs [] = []
removeEqsFrom (e:eqs) eqs' = removeEqsFrom eqs (filter (/=e) eqs')

--
--
-- Unification (according to Martelli and Montanari's)
--
--

mguMM sigma [] = Just sigma
mguMM sigma ((t :-> s):eqs) 
  | s==t = mguMM sigma eqs
mguMM sigma (e@((W x) :-> (W y)):eqs) = mguMM (composeSubst e sigma) (map (applyToEq e) eqs)
mguMM sigma (e@((W x) :-> (V y)):eqs) = mguMM (composeSubst e sigma) (map (applyToEq e) eqs)
mguMM sigma (e@(x@(V _) :-> t):eqs) = 
   if isVarInTerm x t then Nothing else mguMM (composeSubst e sigma) (map (applyToEq e) eqs)
mguMM sigma (e@(x@(W _) :-> t):eqs) = 
   if isVarInTerm x t then Nothing else mguMM (composeSubst e sigma) (map (applyToEq e) eqs)
mguMM sigma ((t :-> (V x)):eqs) = mguMM sigma (((V x) :-> t):eqs)
mguMM sigma ((t :-> (W x)):eqs) = mguMM sigma (((W x) :-> t):eqs)
mguMM sigma (((F f ts) :-> (F g ss)):eqs) 
  | f == g = mguMM sigma ([(ts !! (i-1)) :-> (ss !! (i-1)) | i <- [1..length ts]]++eqs)
mguMM _ _ = Nothing

mgu t s = mguMM [] [s :-> t]

testUnify t s = (mgu t s) /= Nothing


--
--
-- Cut of terms and equations
--
--

infListOfNewVariables = map (('X':).show) [1..]

cutAt k t = fst (cutAt' k t infListOfNewVariables)

cutAt' 0 _ (x:xs) = (W x,xs)
cutAt' k (V x) xs = (V x,xs)
cutAt' k (W x) xs = (W x,xs)
cutAt' k (C a) xs = (C a,xs)
cutAt' k (F f ts) xs = ((F f ss),ys)
   where (ss,ys) = cutListAt (k-1) ts xs
   
cutListAt k [] xs = ([],xs)
cutListAt k (t:ts) xs = (s:ss,ys)
   where (s,zs) = cutAt' k t xs
         (ss,ys) = cutListAt k ts zs

cutEqAt k (t :-> s) = (cutAt k t) :-> (cutAt k s)

cutEqsAt k eqs = map (cutEqAt k) eqs

--
--
--
-- Renaming of variables
--

renWith s (V x) = V (x++s)
renWith s (W x) = W (x++s)
renWith s (F f ts) = F f (map (renWith s) ts)
renWith _ t = t

renRuleWith s (l :-> r) = (renWith s l) :-> (renWith s r)

--
--
-- Making sets of (sorted) ground terms
--
--

makeGSTermsK (funcs,_,repAr,_) 0 = [C c | c <- funcs, ar c==0]
   where ar = toFunc repAr
makeGSTermsK sig@(funcs,_,repAr,repFuncty) k = [C c | c <- funcs, ar c==0] ++ 
   [F f ts | f <- funcs, ar f>0, ts <- makeArgs (tail (functy f)) (ar f)]
   where ar = toFunc repAr
         functy = toFunc repFuncty
         makeArgs sorting arity = combine sorting (replicate arity (makeGSTermsK sig (k-1)))
         combine _ [] = [[]]
         combine (sort:sorts) (xs:xss) = [x:ys | x<-xs, testSort sort x, ys<-combine sorts xss]
         testSort sort (C c) = sort == (head (functy c))
         testSort sort (F f _) = sort == (head (functy f))


extractConstSignature ((funcs,vars,repAr,repFuncty),rules) = (cSymbols,vars,repAr,repFuncty)
   where cSymbols = filter (not.(`elem` dSymbols)) funcs
         dSymbols = map (rootOfTerm.lhsOf) rules
         lhsOf (l :-> _) = l

makeGNFsK' ((funcs,_,ar,_),_) cSymbols dSymbols 0 = 
   [C c | c <- cSymbols, ar c==0]
makeGNFsK' trs@((funcs,vars,ar,functy),rules) cSymbols dSymbols k = 
   [C c | c <- cSymbols, ar c==0] ++ 
   [F f ts | f <- cSymbols, ar f>0, ts <- makeArgs (tail (functy f)) (ar f)] ++
   filter (rootIsNF rules) [F f ts | f <- dSymbols, ar f>0, ts <- makeArgs (tail (functy f)) (ar f)] 
   where makeArgs sorting arity = combine sorting (replicate arity (makeGNFsK' trs cSymbols dSymbols (k-1)))
         combine _ [] = [[]]
         combine (sort:sorts) (xs:xss) = [x:ys | x<-xs, testSort sort x, ys<-combine sorts xss]
         testSort sort (C c) = sort == (head (functy c))
         testSort sort (F f _) = sort == (head (functy f))
         rootIsNF rules t = not (or [testMatch l t | l:->_ <- rules])
         testMatch = testUnify  -- Since we only consider GROUND normal forms!
        
makeGNFsK ((funcs,vars,repAr,repFuncty),rules) k = 
   makeGNFsK' ((funcs,vars,ar,functy),rules) cSymbols dSymbols k
   where ar = toFunc repAr
         functy = toFunc repFuncty
         cSymbols = filter (not.(`elem` dSymbols)) funcs
         dSymbols = map (rootOfTerm.lhsOf) rules
         lhsOf (l :-> _) = l

--
--
-- Equations
--
--

isGroundTerm (C _) = True
isGroundTerm (F _ ts) = and (map isGroundTerm ts)
isGroundTerm _ = False

isGroundEq (t :-> s) = (isGroundTerm t) && (isGroundTerm s)

takeGroundEqs eqs = filter isGroundEq eqs

